unit game;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, StdCtrls, ExtCtrls, ImgList, Grids, Buttons, ComCtrls,
  Math,
  blank_tile;

const
  SQRSZ=30;

  BMP_BLANK=0;
  BMP_STAR=1;
  BMP_W2=2;
  BMP_W3=3;
  BMP_L2=4;
  BMP_L3=5;
  BMP_UNASGN=6;
  BMP_CROSS=7;
  BMP_LTR=8;
  BMP_CX=BMP_LTR+26;
  BMP_DX=BMP_CX+1;
  BMP_EX=BMP_CX+2;
  BMP_NX=BMP_CX+3;
  BMP_RX=BMP_CX+4;
  BMP_SX=BMP_CX+5;
  BMP_TX=BMP_CX+6;
  BMP_UX=BMP_CX+7;
  BMP_ZX=BMP_CX+8;
  BMP_AY=BMP_CX+9;
  BMP_EY=BMP_CX+10;
  BMP_IY=BMP_CX+11;
  BMP_OY=BMP_CX+12;
  BMP_UY=BMP_CX+13;
  BMP_YY=BMP_CX+14;

  END_PASSES=0;
  END_TILES=1;
  END_RESET=2;
  END_ABORT=3;

type
  TGame = class;
  TDragImages = class;
  TBoard = class;
  TPlayer = class;

  TRank = array[1..4] of record
    PlayerNum, Score: Integer;
  end;

  TWindow = class(TForm)
    MainMenu: TMainMenu;
    MenuGame: TMenuItem;
    MenuExit: TMenuItem;
    Info: TMenuItem;
    Oaplikaci: TMenuItem;
    TileImageList: TImageList;
    BoardDrawGrid: TDrawGrid;
    Player1GroupBox: TGroupBox;
    Player1ScoreEdit: TEdit;
    Player1DrawGrid: TDrawGrid;
    Player2GroupBox: TGroupBox;
    Player2ScoreEdit: TEdit;
    Player2DrawGrid: TDrawGrid;
    Player3GroupBox: TGroupBox;
    Player3ScoreEdit: TEdit;
    Player3DrawGrid: TDrawGrid;
    Player4GroupBox: TGroupBox;
    Player4ScoreEdit: TEdit;
    Player4DrawGrid: TDrawGrid;
    PlaceWordButton: TBitBtn;
    BackWordButton: TBitBtn;
    ChallengeWordButton: TBitBtn;
    PassButton: TBitBtn;
    ExchangeButton: TBitBtn;
    MenuNewGame: TMenuItem;
    MenuAction: TMenuItem;
    MenuActionPlace: TMenuItem;
    MenuActionBack: TMenuItem;
    MenuActionChallenge: TMenuItem;
    MenuActionPass: TMenuItem;
    MenuActionExchange: TMenuItem;
    WordsMemo: TMemo;
    HelpMeButton: TBitBtn;
    BagProgressBar: TProgressBar;
    AbortButton: TBitBtn;
    MenuActionHelp: TMenuItem;
    MenuActionAbort: TMenuItem;
    procedure MenuExitClick(Sender: TObject);
    procedure BoardDrawGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FormCreate(Sender: TObject);
    procedure BoardDrawGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure PlayerDrawGridDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure BoardDrawGridDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure PlayerDrawGridDragOver(Sender, Source: TObject; X,
      Y: Integer; State: TDragState; var Accept: Boolean);
    procedure PlayerDrawGridStartDrag(Sender: TObject;
      var DragObject: TDragObject);
    procedure PlayerDrawGridSelectCell(Sender: TObject; ACol,
      ARow: Integer; var CanSelect: Boolean);
    procedure BoardDrawGridDragDrop(Sender, Source: TObject; X,
      Y: Integer);
    procedure BoardDrawGridStartDrag(Sender: TObject;
      var DragObject: TDragObject);
    procedure BoardDrawGridMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PlayerDrawGridMouseDown(Sender: TObject;
      Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure PlayerDrawGridDragDrop(Sender, Source: TObject; X,
      Y: Integer);
    procedure BackWordButtonClick(Sender: TObject);
    procedure PassButtonClick(Sender: TObject);
    procedure MenuNewGameClick(Sender: TObject);
    procedure MenuActionPlaceClick(Sender: TObject);
    procedure MenuActionBackClick(Sender: TObject);
    procedure MenuActionChallengeClick(Sender: TObject);
    procedure MenuActionPassClick(Sender: TObject);
    procedure MenuActionExchangeClick(Sender: TObject);
    procedure OaplikaciClick(Sender: TObject);
    procedure PlaceWordButtonClick(Sender: TObject);
    procedure WordsStringGridSelectCell(Sender: TObject; ACol,
      ARow: Integer; var CanSelect: Boolean);
    procedure ExchangeButtonClick(Sender: TObject);
    procedure ChallengeWordButtonClick(Sender: TObject);
    procedure MenuEndGameClick(Sender: TObject);
    procedure HelpMeButtonClick(Sender: TObject);
    procedure AbortButtonClick(Sender: TObject);
    procedure MenuActionAbortClick(Sender: TObject);
    procedure MenuActionHelpClick(Sender: TObject);
  private
    { Private declarations }
  public
    Game: TGame;
    DragImage: TDragImages;
    procedure AddMemo(Line: String; Sep: Boolean = true);
  end;

  TComputer = class
    constructor Create(Board: TBoard);
    destructor Destroy; override;
  public
    function FindForPlayer(Player: TPlayer): Boolean;
  private
    Board: TBoard;
    Words: TStringList;
    Letters: array[1..7] of Char;
    LettersCount: Integer;
    BestWord: String;
    BestValue: Integer;
    BestR, BestC: Integer;
    BestDir: Boolean;
    BestTiles: array[1..7] of Integer;
    Running: Boolean;
    function FindWord: Boolean;
  end;

  TTile = class
    constructor Create(ltr: Char);
  public
    Blank: Boolean;
    Letter: Char;
    function Value: Integer;
  end;

  TSquare = class
    constructor Create(Board: TBoard; Player: TPlayer; r, c, wm, lm: Integer);
    destructor Destroy;override;
  public
    Row: Integer;
    Column: Integer;
    WordMul: Integer;
    LetterMul: Integer;
    Tile: TTile;
    ThisTurn: Boolean;
    Empty: Boolean;
    Bitmap: TBitmap;
    function BitmapIndex: Integer;
    function PlaceTile(Tile: TTile): Boolean;
    procedure RemoveTile(Method: Boolean=true);
    function Mark: Boolean;
    function AssignLetter(Letter: Char): Boolean;
  private
    Board: TBoard;
    Player: TPlayer;
    Marked: Boolean;
  end;

  TBoard = class
    constructor Create(Game: TGame);
    destructor Destroy;override;
  public
    Squares: array[1..15, 1..15] of TSquare;
    procedure AddSquare(Square: TSquare);
    procedure RemoveSquare(Square: TSquare);
    procedure Pass(ClearOnly: Boolean = false);
    procedure Draw;
    procedure Clear;
    function PlaceWord: Integer;
    function CanChallenge: Boolean;
    procedure Mark(Square: TSquare; Mark: Boolean);
    function ChallengeWord: Integer;
  private
    Game: TGame;
    ThisTurn: TList;
    LastTurn: TList;
    Marked: TList;
    OnBoard: Integer;
    LastScore: Integer;
    procedure BackWord(Num: Integer);
    function ValidState: Boolean;
    function OnlyOneRowCol: Boolean;
    function Score(var List: TList): Integer;
  end;

  TBag = class
    constructor Create;
    destructor Destroy; override;
  public
    Tiles: TList;
    function TilesLeft: Integer;
    procedure Refill;
    procedure Shuffle;
    function FetchTiles(Count: Integer; _tiles: TList; Upd: Boolean = true): Integer;
    procedure Draw;
    function ExchangeTiles(TL: TList): Boolean;
    function CanExchange: Boolean;
    function ReturnTiles(TL: TList): Integer;
  end;

  TPlayer = class
    constructor Create(Game: TGame; Name: String; Num: Integer; Computer: Boolean=false);
    destructor Destroy; override;
  public
    Game: TGame;
    Squares: array[1..7] of TSquare;
    Num: Integer;
    Name: String;
    Score: Integer;
    Computer: Boolean;
    function GetTiles: Integer;
    procedure Draw;
    procedure BackWord;
    function BackTile(Tile: TTile): Boolean;
    procedure Activate;
    procedure Deactivate;
    function ExchangeTiles: Boolean;
    function EndPlayer: Integer;
  private
    GroupBox: TGroupBox;
    DrawGrid: TDrawGrid;
    ScoreEdit: TEdit;
    TilesNum: Integer;
  end;

  TDragImages = class(TDragControlObject)
    constructor Create(AControl: TControl);override;
    destructor Destroy; override;
  public
    Square: TSquare;
    function GetDragImages: TDragImageList; override;
    function GetDragCursor(Accepted: Boolean; X, Y: Integer): TCursor; override;
  end;

  TGame = class
    constructor Create;
  public
    Board: TBoard;
    Bag: TBag;
    Players: array[1..4] of TPlayer;
    Playing: Boolean;
    Abort: Boolean;
    curPlayer, numPlayers: Integer;
    function StartGame: Boolean;
    procedure EndGame(Reason: Integer);
    function MapPlayerDrawGrid(Sender: TObject): Integer;
    function CurDrawGrid: TDrawGrid;
    procedure Pass(IncOnly: Boolean = false);
    procedure NoPass;
    procedure NextPlayer;
    function PlaceWord: Integer;
    function Player: TPlayer;
    function ExchangeTiles: Boolean;
    function ChallengeWord: Integer;
    function PrevPlayer: TPlayer;
    procedure HelpPlayer;
  private
    numPasses: Integer;
    Comp: TComputer;
    procedure Rank(var r: TRank);
    procedure GetPlayers;
    procedure PlayComputer;
  end;

  TWord = class
    constructor Create(Text: String; Score: Integer);
  public
    Text: String[15];
    Score: Integer;
    function GetText: String;
  end;

var
  Window: TWindow;
  s: String;
  TilesOrder: array[32..255] of Byte;

procedure InitSort;

implementation

uses players;

{$R *.dfm}

// **********************************************
// ******************** General ********************
// **********************************************

procedure SetControls(c: TWinControl);
var
  i: Integer;
begin
  c.ControlStyle:=c.ControlStyle+[csDisplayDragImage];
  for i:=0 to c.ControlCount-1 do
   if c is TWinControl then SetControls(TWinControl(c.Controls[i]));
end;

function SortTiles(Item1, Item2: Pointer): Integer;
begin
  Result:=0;
  if (Item1=nil) or (Item2=nil) then exit;
  Result:=TilesOrder[ord(TTile(Item1).Letter)]-TilesOrder[ord(TTile(Item2).Letter)];
end;

procedure InitSort;
var
  i: Integer;
begin
  for i:=32 to 255 do TilesOrder[i]:=0;

  TilesOrder[ord('A')]:=1;
  TilesOrder[ord('')]:=2;
  TilesOrder[ord('B')]:=3;
  TilesOrder[ord('C')]:=4;
  TilesOrder[ord('')]:=5;
  TilesOrder[ord('D')]:=6;
  TilesOrder[ord('')]:=7;
  TilesOrder[ord('E')]:=8;
  TilesOrder[ord('')]:=9;
  TilesOrder[ord('')]:=10;
  TilesOrder[ord('F')]:=11;
  TilesOrder[ord('G')]:=12;
  TilesOrder[ord('H')]:=13;
  TilesOrder[ord('I')]:=14;
  TilesOrder[ord('')]:=15;
  TilesOrder[ord('J')]:=16;
  TilesOrder[ord('K')]:=17;
  TilesOrder[ord('L')]:=18;
  TilesOrder[ord('M')]:=19;
  TilesOrder[ord('N')]:=20;
  TilesOrder[ord('')]:=21;
  TilesOrder[ord('O')]:=22;
  TilesOrder[ord('')]:=23;
  TilesOrder[ord('P')]:=24;
  TilesOrder[ord('Q')]:=25;
  TilesOrder[ord('R')]:=26;
  TilesOrder[ord('')]:=27;
  TilesOrder[ord('S')]:=28;
  TilesOrder[ord('')]:=29;
  TilesOrder[ord('T')]:=30;
  TilesOrder[ord('')]:=31;
  TilesOrder[ord('U')]:=32;
  TilesOrder[ord('')]:=33;
  TilesOrder[ord('')]:=34;
  TilesOrder[ord('V')]:=35;
  TilesOrder[ord('W')]:=36;
  TilesOrder[ord('X')]:=37;
  TilesOrder[ord('Y')]:=38;
  TilesOrder[ord('')]:=39;
  TilesOrder[ord('Z')]:=40;
  TilesOrder[ord('')]:=41;
  TilesOrder[ord(' ')]:=42;
end;

// **********************************************
// ******************** TBag ********************
// **********************************************

constructor TBag.Create;
begin
  inherited Create;
  Tiles:=TList.Create;
  Refill;
end;

destructor TBag.Destroy;
begin
  Tiles.Clear;
  Tiles.Free;
  inherited Destroy;
end;

procedure TBag.Refill;
var
  i: Integer;
begin
  Tiles.Clear;
  for i:=1 to 5 do Tiles.Add(TTile.Create('A'));
  for i:=1 to 2 do Tiles.Add(TTile.Create(''));
  for i:=1 to 2 do Tiles.Add(TTile.Create('B'));
  for i:=1 to 3 do Tiles.Add(TTile.Create('C'));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 3 do Tiles.Add(TTile.Create('D'));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 5 do Tiles.Add(TTile.Create('E'));
  for i:=1 to 2 do Tiles.Add(TTile.Create(''));
  for i:=1 to 2 do Tiles.Add(TTile.Create(''));
  for i:=1 to 1 do Tiles.Add(TTile.Create('F'));
  for i:=1 to 1 do Tiles.Add(TTile.Create('G'));
  for i:=1 to 3 do Tiles.Add(TTile.Create('H'));
  for i:=1 to 4 do Tiles.Add(TTile.Create('I'));
  for i:=1 to 3 do Tiles.Add(TTile.Create(''));
  for i:=1 to 2 do Tiles.Add(TTile.Create('J'));
  for i:=1 to 3 do Tiles.Add(TTile.Create('K'));
  for i:=1 to 3 do Tiles.Add(TTile.Create('L'));
  for i:=1 to 3 do Tiles.Add(TTile.Create('M'));
  for i:=1 to 5 do Tiles.Add(TTile.Create('N'));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 6 do Tiles.Add(TTile.Create('O'));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 3 do Tiles.Add(TTile.Create('P'));
  for i:=1 to 3 do Tiles.Add(TTile.Create('R'));
  for i:=1 to 2 do Tiles.Add(TTile.Create(''));
  for i:=1 to 4 do Tiles.Add(TTile.Create('S'));
  for i:=1 to 2 do Tiles.Add(TTile.Create(''));
  for i:=1 to 4 do Tiles.Add(TTile.Create('T'));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 3 do Tiles.Add(TTile.Create('U'));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 4 do Tiles.Add(TTile.Create('V'));
  for i:=1 to 1 do Tiles.Add(TTile.Create('X'));
  for i:=1 to 2 do Tiles.Add(TTile.Create('Y'));
  for i:=1 to 2 do Tiles.Add(TTile.Create(''));
  for i:=1 to 2 do Tiles.Add(TTile.Create('Z'));
  for i:=1 to 1 do Tiles.Add(TTile.Create(''));
  for i:=1 to 2 do Tiles.Add(TTile.Create(' '));
  Shuffle;
  Draw;
end;

procedure TBag.Shuffle;
var
  i: Integer;
begin
  for i:=0 to Tiles.Count-1 do Tiles.Exchange(i, i+Random(Tiles.Count-i));
end;

function TBag.TilesLeft: Integer;
begin
  Result:=Tiles.Count;
end;

function TBag.FetchTiles(Count: Integer; _tiles: TList; Upd: Boolean = true): Integer;
var
  sf: Integer;
begin
  sf:=0;
  if (Count<0) or (Count>7) then sf:=-1
  else while (Count>0) do
  begin
    Dec(Count);
    if Tiles.Count=0 then break;
    _tiles.Add(Tiles.Extract(Tiles.Last));
    Inc(sf);
  end;
  Result:=sf;
  if Upd then Draw;
end;

procedure TBag.Draw;
begin
  Window.BagProgressBar.Position:=Tiles.Count;
  Window.BagProgressBar.Invalidate;
end;

function TBag.ExchangeTiles(TL: TList): Boolean;
var
  EL: TList;
begin
  Result:=false;
  if not CanExchange then exit;
  if (TL.Count<1) or (TL.Count>7) then exit;

  EL:=TList.Create;
  FetchTiles(TL.Count, EL, false);
  ReturnTiles(TL);
  TL.Assign(EL);
  El.Clear;
  EL.Free;

  Window.AddMemo(IntToStr(Window.Game.curPlayer)+ ': mnil '+IntToStr(TL.Count));
  Result:=true;
  Draw;
end;

function TBag.CanExchange: Boolean;
begin
  Result:=Tiles.Count>=7;
end;

function TBag.ReturnTiles(TL: TList): Integer;
begin
  Result:=TL.Count;
  while TL.Count>0 do Tiles.Add(TL.Extract(TL.Last));
  TL.Clear;
  Shuffle;
end;

// ************************************************
// ******************** TBoard ********************
// ************************************************

constructor TBoard.Create(Game: TGame);
var
  i,j: Integer;
  lm, wm: Integer;
begin
  inherited Create;

  Self.Game:=Game;
  ThisTurn:=TList.Create;
  LastTurn:=TList.Create;
  Marked:=TList.Create;
  OnBoard:=0;

  with Window.BoardDrawGrid.Canvas do
  begin
    Brush.Color:=clGray;
    Font.Name:='Verdana';
    Font.Size:=15;
    Font.Style:=[fsBold];
  end;

  for i:=1 to 15 do for j:=1 to 15 do
  begin
    lm:=1;
    wm:= 1;
    if (((i=1) and ((j=1) or (j=8) or (j=15))) or ((i=8) and ((j=1) or (j=15)))
     or ((i=15) and ((j=1) or (j=8) or (j=15)))) then wm:=3
    else if (((((i>=2) and (i<=5)) or ((i>=11) and (i<=14)) or (i=8)) and ((j=i) or (j=(16-i)))) or ((i=8) and (j=8))) then wm:=2
    else if (((i=2) or (i=6) or (i=10) or (i=14)) and ((j=2) or (j=6) or (j=10) or (j=14))) then lm:=3
    else if ((((i=1) or (i=8) or (i=15)) and ((j=4) or (j=12))) or (((i=3) or (i=7) or (i=9) or (i=13))
     and ((j=3) or (j=7) or (j=9) or (j=13))) or (((i=4) or (i=12)) and ((j=1) or (j=8) or (j=15)))) then lm:=2;

    Squares[i,j]:=TSquare.Create(Self, nil, i, j, wm, lm);
  end;
  Draw;
end;

destructor TBoard.Destroy;
begin
  Marked.Clear;
  Marked.Free;
  LastTurn.Clear;
  LastTurn.Free;
  ThisTurn.Clear;
  ThisTurn.Free;
  inherited Destroy;
end;

procedure TBoard.Draw;
begin
  if Game.curPlayer=0 then exit;
  if Game.Playing and not Game.Player.Computer and (ThisTurn.Count > 0) then Window.BackWordButton.Enabled:=true
  else Window.BackWordButton.Enabled:=false;

  if Game.Playing and not Game.Player.Computer and (ThisTurn.Count > 0) and (OnBoard >= 2) and not Squares[8, 8].Empty
   and ValidState and OnlyOneRowCol then Window.PlaceWordButton.Enabled:=true
  else Window.PlaceWordButton.Enabled:=false;

  Window.ExchangeButton.Enabled:=Game.Playing and not Game.Player.Computer and (Marked.Count > 0);
  Window.ChallengeWordButton.Enabled:=Game.Playing and not Game.Player.Computer and not Game.PrevPlayer.Computer and (LastTurn.Count > 0);
  Window.HelpMeButton.Enabled:=Game.Playing and not Game.Player.Computer and (ThisTurn.Count=0) and (Marked.Count=0);
  Window.PassButton.Enabled:=Game.Playing and not Game.Player.Computer;

  Window.BoardDrawGrid.Invalidate;
end;

function TBoard.ValidState: Boolean;
var
  i, j: Integer;
begin
  Result:=true;
  for i:=1 to 15 do for j:=1 to 15 do
  if not Squares[i, j].Empty and ((Squares[i, j].Tile.Letter = ' ') or
   (((i<>8) or (j<>8)) and
    (((i=1) or Squares[i-1, j].Empty) and ((i=15) or Squares[i+1, j].Empty)
    and ((j=1) or Squares[i, j-1].Empty) and ((j=15) or Squares[i, j+1].Empty)))) then
  begin
    Result:=false;
    exit;
  end;
end;

function TBoard.OnlyOneRowCol: Boolean;
var
  i: Integer;
  r, c, rx, cx: Integer;
  minr, maxr, minc, maxc: Integer;
  e: Boolean;
begin
  Result:=true;
  if ThisTurn.Count<2 then exit;

  r:=0;
  c:=0;
  minr:=16; maxr:=0; minc:=16; maxc:=0;

  for i:=0 to ThisTurn.Count-1 do
  begin
    rx:=TSquare(ThisTurn[i]).Row;
    cx:=TSquare(ThisTurn[i]).Column;
    if r=0 then r:=rx;
    if c=0 then c:=cx;
    if rx < minr then minr:=rx;
    if rx > maxr then maxr:=rx;
    if cx < minc then minc:=cx;
    if cx > maxc then maxc:=cx;

    if (r=rx) and (c>0) and (c<>cx) then c:=-1
    else if (c=cx) and (r>0) and (r<>rx) then r:=-1
    else if (c>0) and (r>0) and (c<>cx) and (r<>rx) then begin Result:=false; exit; end;

    if ((r<>rx) and (c=-1)) or ((c<>cx) and (r=-1)) then
    begin Result:=false; exit; end;
  end;

  e:=Squares[8, 8].ThisTurn;
  if r=-1 then
  begin
    if (minr>1) and not Squares[minr-1, c].Empty then Dec(minr);
    if (maxr<15) and not Squares[maxr+1, c].Empty then Inc(maxr);

    for i:=minr to maxr do
    begin
      if Squares[i, c].Empty then Result:=false;
      if not Squares[i, c].ThisTurn or
       ((c>1) and not (Squares[i, c-1].ThisTurn or Squares[i, c-1].Empty)) or
       ((c<15) and not (Squares[i, c+1].ThisTurn or Squares[i, c+1].Empty)) then e:=true;
    end;
  end;
  if c=-1 then
  begin
    if (minc>1) and not Squares[r, minc-1].Empty then Dec(minc);
    if (maxc<15) and not Squares[r, maxc+1].Empty then Inc(maxc);
    for i:=minc to maxc do
    begin
      if Squares[r, i].Empty then Result:=false;
      if not Squares[r, i].ThisTurn or
       ((r>1) and not (Squares[r-1, i].ThisTurn or Squares[r-1, i].Empty)) or
       ((r<15) and not (Squares[r+1, i].ThisTurn or Squares[r+1, i].Empty)) then e:=true;
    end;
  end;
  Result:=Result and e;
end;

procedure TBoard.BackWord(Num: Integer);
begin
  while ThisTurn.Count > 0 do
  if Game.Players[Num].BackTile(TSquare(ThisTurn.Last).Tile) then
    TSquare(ThisTurn.Last).RemoveTile;
  Draw;
  Window.BoardDrawGrid.Invalidate;
  Game.CurDrawGrid.Invalidate;
end;

procedure TBoard.Pass(ClearOnly: Boolean = false);
begin
  BackWord(Game.curPlayer);
  LastTurn.Clear;
  Marked.Clear;
  if not ClearOnly then Window.AddMemo(IntToStr(Game.curPlayer)+ ': ekal');
end;

procedure TBoard.Clear;
var
  i, j: Integer;
begin
  for i:=1 to 15 do  for j:=1 to 15 do
  if Squares[i, j].Tile <> nil then
    Squares[i, j].RemoveTile;
  ThisTurn.Clear;
  LastTurn.Clear;
  Marked.Clear;
  Draw;
  Window.BoardDrawGrid.Invalidate;
end;

procedure TBoard.AddSquare(Square: TSquare);
begin
  ThisTurn.Add(Square);
  Inc(OnBoard);
  LastTurn.Clear;
  Draw;
end;

procedure TBoard.RemoveSquare(Square: TSquare);
begin
  ThisTurn.Remove(Square);
  Dec(OnBoard);
  Draw;
end;

function TBoard.PlaceWord: Integer;
var
  l: TList;
  i: Integer;
begin
  Result:=0;
  if (ThisTurn.Count = 0) or (OnBoard < 2) or Squares[8, 8].Empty
   or not ValidState or not OnlyOneRowCol then exit;

  // generate new words
  l:=TList.Create;
  LastScore:=Score(l);
  Result:=LastScore;
  for i:=0 to l.Count-1 do
    Window.AddMemo(IntToStr(Game.curPlayer)+ ': ' + TWord(l[i]).GetText, i=(l.Count-1));
  l.Clear;
  l.Free;

  for i:=0 to ThisTurn.Count-1 do TSquare(ThisTurn[i]).ThisTurn:=false;
  LastTurn.Clear;
  LastTurn.Assign(ThisTurn);
  ThisTurn.Clear;
  Marked.Clear;
  Draw;
end;

function TBoard.Score(var List: TList): Integer;
var
  sc: Integer;
  i, j, k: Integer;
  sw, wm, lm: Integer;
  s: String;
begin
  sc:=0;

  for i:=1 to 15 do
  begin
    j:=1;
    while (j<=15) and not Squares[i, j].ThisTurn do Inc(j);
    if j>15 then continue;
    while (j>1) and not Squares[i, j-1].Empty do Dec(j);
    // j is start of word
    k:=j;
    while (k<15) and not Squares[i, k+1].Empty do Inc(k);
    // k is end of word
    if (j=k) then continue; // no horizontal word

    wm:=1;
    sw:=0;
    s:='';
    for j:=j to k do
    begin
      if Squares[i, j].ThisTurn then
      begin
        wm:=wm * Squares[i, j].WordMul;
        lm:=Squares[i, j].LetterMul;
      end
      else lm:=1;
      sw:=sw + Squares[i, j].Tile.Value * lm;
      s:=s + Squares[i, j].Tile.Letter;
    end;
    sw:= sw * wm;
    sc:= sc + sw;
    List.Add(TWord.Create(s, sw));
  end;

  for i:=1 to 15 do
  begin
    j:=1;
    while (j<=15) and not Squares[j, i].ThisTurn do Inc(j);
    if j>15 then continue;
    while (j>1) and not Squares[j-1, i].Empty do Dec(j);
    // j is start of word
    k:=j;
    while (k<15) and not Squares[k+1, i].Empty do Inc(k);
    // k is end of word
    if (j=k) then continue; // no vertical word

    wm:=1;
    sw:=0;
    s:='';
    for j:=j to k do
    begin
      if Squares[j, i].ThisTurn then
      begin
        wm:=wm * Squares[j, i].WordMul;
        lm:=Squares[j, i].LetterMul;
      end
      else lm:=1;
      sw:=sw + Squares[j, i].Tile.Value * lm;
      s:=s + Squares[j, i].Tile.Letter;
    end;
    sw:= sw * wm;
    sc:= sc + sw;
    List.Add(TWord.Create(s, sw));
  end;

  if ThisTurn.Count=7 then sc:=sc+50;
  Result:=sc;
end;

function TBoard.CanChallenge: Boolean;
begin
  Result:=LastTurn.Count>0;
end;

procedure TBoard.Mark(Square: TSquare; Mark: Boolean);
begin
  if Square.Tile=nil then exit;
  if Mark then Marked.Add(Square)
  else Marked.Remove(Square);
  Draw;
end;

function TBoard.ChallengeWord: Integer;
var
  w: TWord;
  i: Integer;
begin
  if not CanChallenge or
   (Application.MessageBox('Souhlas s odmtnutm pedchozho slova (slov)?',
   PChar('Otzka na hre '+Game.PrevPlayer.Name), MB_YESNO) = IDNO) then
  begin Result:=-1; exit; end;

  w:=TWord.Create('Odmtnut', -LastScore);
  Window.AddMemo(IntToStr(Game.PrevPlayer.Num)+ ': ' + w.GetText);
  w.Free;

  for i:=0 to LastTurn.Count-1 do TSquare(LastTurn[i]).RemoveTile(false);
  LastTurn.Clear;

  Result:=LastScore;
  Window.BoardDrawGrid.Invalidate;
  Draw;
end;

// ***************************************************
// ******************** TComputer ********************
// ***************************************************

constructor TComputer.Create(Board: TBoard);
begin
  inherited Create;
  Self.Board:=Board;
  LettersCount:=0;
  Running:=false;

  Words:=TStringList.Create;
  Words.CaseSensitive:=false;
  try
    Words.LoadFromFile('slovnik.txt');
  except
    on EFOpenError do ShowMessage('Nenael jsem slovnik.txt');
  end;
  Words.Sort;
end;

destructor TComputer.Destroy;
begin
  Words.Clear;
  Words.Free;
  inherited Destroy;
end;

function TComputer.FindForPlayer(Player: TPlayer): Boolean;
var
  i: Integer;
begin
  LettersCount:=0;
  for i:=1 to 7 do if Player.Squares[i].Tile <> nil then
  begin
    Inc(LettersCount);
    Letters[LettersCount]:=Player.Squares[i].Tile.Letter;
  end;
  Running:=true;
  Window.AbortButton.Enabled:=true;
  Window.PassButton.Enabled:=false;
  Result:=FindWord;
  if not Player.Computer then Window.PassButton.Enabled:=true;
  Window.AbortButton.Enabled:=false;
  Running:=false;
end;

function TComputer.FindWord: Boolean;
var
  i, j, k: Integer;
  m: array[1..5] of Integer;
  m1, m2, m3, m4, m5: Integer;
  d: Integer;
  st: String[5];
  t: TTile;

procedure Calc(w: String; r, c: Integer; dir: Boolean=false);
var
  i: Integer;
  v: Integer;
  wm: Integer;
begin
  v:=0;
  wm:=1;

  for i:=0 to Length(w)-1 do
  if dir then
  begin
    if Board.Squares[r+i, c].Tile<>nil then Inc(v, Board.Squares[r+i, c].Tile.Value)
    else
    begin
      t.Letter:=w[i+1];
      Inc(v, t.Value * Board.Squares[r+i, c].LetterMul);
      wm:=wm*Board.Squares[r+i, c].WordMul;
    end;
  end
  else
  begin
    if Board.Squares[r, c+i].Tile<>nil then Inc(v, Board.Squares[r, c+i].Tile.Value)
    else
    begin
      t.Letter:=w[i+1];
      Inc(v, t.Value * Board.Squares[r, c+i].LetterMul);
      wm:=wm*Board.Squares[r, c+i].WordMul;
    end;
  end;

  v:=v*wm;
  if v>BestValue then
  begin
    BestWord:=w;
    BestValue:=v;
    BestR:=r;
    BestC:=c;
    BestDir:=dir;
    for i:=1 to 5 do BestTiles[i]:=m[i];
  end;
end;

function Check(ll, lr: Integer): Boolean;
var
  w: array[1..15] of char;
  w2: String[15];
  p, q: Integer;
  r, c, ce: Integer;
  ind: Integer;

begin
  Result:=false;

  // vodorovne
  if Board.Squares[i, j].Tile<>nil then w[j]:=Board.Squares[i, j].Tile.Letter;

  r:=i; c:=j; ce:=j;
  q:=ll;
  for p:=j-1 downto 1 do
  begin
    if ((i>1) and not Board.Squares[i-1, p].Empty) or
     ((i<15) and not Board.Squares[i+1, p].Empty) then exit
    else if not Board.Squares[i, p].Empty then w[p]:=Board.Squares[i, p].Tile.Letter
    else
    begin
      if q=0 then break;
      w[p]:=st[q];
      Dec(q);
    end;
    c:=p;
  end;
  if q>0 then exit;

  if (ce-c)>4 then exit;

  q:=1;
  for p:=j+1 to 15 do
  begin
    if ((i>1) and not Board.Squares[i-1, p].Empty) or
     ((i<15) and not Board.Squares[i+1, p].Empty) then exit
    else if not Board.Squares[i, p].Empty then w[p]:=Board.Squares[i, p].Tile.Letter
    else
    begin
      if q>lr then break;
      w[p]:=st[ll+q];
      Inc(q);
    end;
    ce:=p;
  end;
  if q<=lr then exit;

  if (ce-c)>4 then exit;

  if Board.Squares[i, j].Tile<>nil then w2:=Copy(w, c, ce-c+1)
  else w2:=Copy(w, c, j-c)+Copy(w, j+1, ce-j);

  if Words.Find(w2, ind) then Calc(w2, r, c);
end;

function CheckS(ll, lr: Integer): Boolean;
var
  w: array[1..15] of char;
  w2: String[15];
  p, q: Integer;
  r, c, re: Integer;
  ind: Integer;

begin
  Result:=false;

  // svisle
  if Board.Squares[i, j].Tile=nil then exit;
  w[i]:=Board.Squares[i, j].Tile.Letter;

  r:=i; c:=j; re:=i;
  q:=ll;
  for p:=i-1 downto 1 do
  begin
    if ((j>1) and not Board.Squares[p, j-1].Empty) or
     ((j<15) and not Board.Squares[p, j+1].Empty) then exit
    else if not Board.Squares[p, j].Empty then w[p]:=Board.Squares[p, j].Tile.Letter
    else
    begin
      if q=0 then break;
      w[p]:=st[q];
      Dec(q);
    end;
    r:=p;
  end;
  if q>0 then exit;

  if (re-r)>4 then exit;

  q:=1;
  for p:=i+1 to 15 do
  begin
    if ((j>1) and not Board.Squares[p, j-1].Empty) or
     ((j<15) and not Board.Squares[p, j+1].Empty) then exit
    else if not Board.Squares[p, j].Empty then w[p]:=Board.Squares[p, j].Tile.Letter
    else
    begin
      if q>lr then break;
      w[p]:=st[ll+q];
      Inc(q);
    end;
    re:=p;
  end;
  if q<=lr then exit;

  if (re-r)>4 then exit;
  w2:=Copy(w, r, re-r+1);

  if Words.Find(w2, ind) then Calc(w2, r, c, true);
end;

begin
  // pro kazde obsazene misto
  // vodorovne i svisle
  // zkusim kombinace pismen pred a za nim

  BestWord:='';
  BestValue:=0;
  t:=TTile.Create('A');
  Result:=false;

  if Board.Squares[8, 8].Empty then Board.Squares[8, 8].Empty:=false;

  for i:=1 to 15 do for j:=1 to 15 do
  if not Board.Squares[i, j].Empty then
  begin
    Application.ProcessMessages;

    for m1:=0 to LettersCount do
    for m2:=0 to LettersCount do
    for m3:=0 to LettersCount do
    for m4:=0 to LettersCount do
    for m5:=1 to LettersCount do
    begin
      if ((m1>0) and ((m1=m2) or (m1=m3) or (m1=m4) or (m1=m5))) or
       ((m2>0) and ((m2=m3) or (m2=m4) or (m2=m5))) or
       ((m3>0) and ((m3=m4) or (m3=m5))) or
       ((m4>0) and ((m4=m5))) then continue;

      m[1]:=m1; m[2]:=m2; m[3]:=m3; m[4]:=m4; m[5]:=m5;

      d:=0;
      for k:=1 to 5 do if m[k]>0 then begin Inc(d); st[d]:=Letters[m[k]]; end;
      SetLength(st, d);

//      if (m3=0) and (m4=0) and (m5=1) then Application.ProcessMessages;
      if Window.Game.Abort then exit;

      for k:=0 to d do
      begin
        Check(k, d-k);
        CheckS(k, d-k);
      end;
    end;
  end;
  Result:=BestWord <> '';
  if Board.Squares[8, 8].Tile=nil then Board.Squares[8, 8].Empty:=true;
  t.Free;
end;

// *****************************************************
// ******************** TDragImages ********************
// *****************************************************

constructor TDragImages.Create(AControl: TControl);
begin
  inherited Create(AControl);
end;

destructor TDragImages.Destroy;
begin
  inherited Destroy;
end;

function TDragImages.GetDragCursor(Accepted: Boolean; X, Y: Integer): TCursor;
begin
  Result := crNone;
end;

function TDragImages.GetDragImages: TDragImageList;
begin
  if Window.DragImage.Square<>nil then
  begin
    with Window.TileImageList do SetDragImage(Window.DragImage.Square.BitmapIndex, Width div 2, Height div 2);
    Result:=Window.TileImageList;
  end
  else Result:=nil;
end;

// ***********************************************
// ******************** TGame ********************
// ***********************************************

constructor TGame.Create;
begin
  inherited Create;
  Randomize;
  Abort:=false;
  Board:=TBoard.Create(Self);
  Bag:=TBag.Create;
  Comp:=TComputer.Create(Board);

  curPlayer:=1;
  numPlayers:=0;
  numPasses:=0;
end;

function TGame.StartGame: Boolean;
begin
  GetPlayers;
//  numPlayers:=0;
  Result:=false;
  if (numPlayers < 2) or not Playing then exit;

  numPasses:=0;
  curPlayer:=0;
//  Playing:=true;
  Abort:=false;

  Board.Clear;
//  Bag.Refill;

  Result:=true;

  Window.WordsMemo.Text:='';

  Window.PassButton.Enabled:=true;
  Window.Invalidate;
//  Players[curPlayer].Activate;
  NextPlayer;
end;

procedure TGame.EndGame(Reason: Integer);
var
  i: Integer;
  s: Integer;
  p: Integer;
  x: Integer;
  r: TRank;
  str: String;
begin
  if not Playing then exit;

  Players[curPlayer].Deactivate;
  Board.Draw;

  Window.BackWordButton.Enabled:=false;
  Window.PlaceWordButton.Enabled:=false;
  Window.ExchangeButton.Enabled:=false;
  Window.ChallengeWordButton.Enabled:=false;
  Window.PassButton.Enabled:=false;
  Window.HelpMeButton.Enabled:=false;

  Playing:=false;
  if (Reason = END_RESET) or (Reason = END_ABORT) then exit;

  p:=-1;
  s:=0;
  for i:=1 to numPlayers do
  begin
    x:=Players[i].EndPlayer;
    if (x=-1) then p:=i
    else Inc(s, x);
  end;

  if (p<>-1) then Inc(Players[p].Score, s);
  for i:=1 to numPlayers do Players[i].Draw;

  Rank(r);

  case Reason of
    END_TILES: str:='Nejsou dal kameny.';
    END_PASSES: str:='Dv kola bez poloen kamen.';
  end;
  str:=str+chr(13)+chr(13)+'Vsledn poad hr je nsledovn:'+chr(13);
  for i:=1 to numPlayers do
   str:=str+chr(13)+IntToStr(i)+'. msto: '+Players[r[i].PlayerNum].Name+', '+
    IntToStr(r[i].Score)+' bod';

  Application.MessageBox(PChar(str), 'Konec hry', MB_OK);
end;

function TGame.MapPlayerDrawGrid(Sender: TObject): Integer;
var
  i: Integer;
begin
  Result:=0;
  for i:=1 to numPlayers do if Sender=Players[i].DrawGrid then Result:=i;
end;

function TGame.CurDrawGrid: TDrawGrid;
begin
  if numPlayers=0 then Result:=nil
  else Result:=Players[curPlayer].DrawGrid;
end;

procedure TGame.Pass(IncOnly: Boolean = false);
begin
  Inc(numPasses);

  if numPasses = 2*numPlayers then
  begin
    EndGame(END_PASSES);
  end
  else if not IncOnly then
  begin
    Board.Pass;
    NextPlayer;
  end;
end;

procedure TGame.NoPass;
begin
  numPasses:=0;
  NextPlayer;
end;

procedure TGame.NextPlayer;
begin
  if not Playing then exit;

  if curPlayer>0 then Players[curPlayer].Deactivate;
  if Abort then EndGame(END_ABORT)
   else if (curPlayer>0) and (Player.TilesNum=0) then EndGame(END_TILES)
  else
  begin
    Inc(curPlayer);
    if curPlayer > numPlayers then curPlayer:=1;
    Players[curPlayer].Activate;
    Board.Draw;
    if Players[curPlayer].Computer then PlayComputer;
  end;
end;

function TGame.PlaceWord: Integer;
begin
  Result:=Board.PlaceWord;
  if Result >=0 then
  begin
    numPasses:=0;
    Player.Score:=Player.Score + Result;
    NextPlayer;
  end;
end;

function TGame.Player: TPlayer;
begin
  if curPlayer > 0 then Result:=Players[curPlayer]
  else Result:=nil;
end;

function TGame.ExchangeTiles: Boolean;
begin
  Result:=Player.ExchangeTiles;
  if Result then
  begin
    Pass(true);
    Board.Pass(true);
    NextPlayer;
  end;
end;

function TGame.ChallengeWord: Integer;
begin
  Result:=Board.ChallengeWord;
  if Result >=0 then
  begin
    PrevPlayer.Score:=PrevPlayer.Score - Result;
    PrevPlayer.Draw;
    Pass(true);
  end;
end;

function TGame.PrevPlayer: TPlayer;
begin
  if curPlayer=1 then Result:=Players[numPlayers]
  else Result:=Players[curPlayer-1];
end;

procedure TGame.Rank(var r: TRank);
var
  i, j: Integer;
  x: Integer;
begin
  for i:=1 to numPlayers do
  begin
    r[i].PlayerNum:=i;
    r[i].Score:=Players[i].Score;
  end;

  for i:=2 to numPlayers do
  for j:=numPlayers downto i do
  if r[j-1].Score < r[j].Score then
  begin
    x:=r[j-1].Score; r[j-1].Score:=r[j].Score; r[j].Score:=x;
    x:=r[j-1].PlayerNum; r[j-1].PlayerNum:=r[j].PlayerNum; r[j].PlayerNum:=x;
  end;
end;

procedure TGame.GetPlayers;
var
  i: Integer;
begin
  if PlayersForm.ShowModal > 0 then
  with PlayersForm do
  begin
    for i:=1 to numPlayers do Players[i].Free;
    numPlayers:=NumPlayersUpDown.Position;

    Bag.Refill;
    Playing:=true;

    if numPlayers >=1 then Players[1]:=TPlayer.Create(Self, Player1Edit.Text, 1, Player1CheckBox.Checked);
    if numPlayers >=2 then Players[2]:=TPlayer.Create(Self, Player2Edit.Text, 2, Player2CheckBox.Checked);
    if numPlayers >=3 then Players[3]:=TPlayer.Create(Self, Player3Edit.Text, 3, Player3CheckBox.Checked);
    if numPlayers >=4 then Players[4]:=TPlayer.Create(Self, Player4Edit.Text, 4, Player4CheckBox.Checked);
  end;
end;

procedure TGame.HelpPlayer;
var
  b: Boolean;
begin
  if not Playing then exit;
  b:=Comp.FindForPlayer(Player);
  if Abort then EndGame(END_ABORT)
  else if b then
   ShowMessage(Comp.BestWord+', '+chr(Comp.BestC+ord('A')-1)+' '+IntToStr(Comp.BestR))
  else ShowMessage('Nenael jsem dn slovo.');
end;

procedure TGame.PlayComputer;
var
  i, j, k, l: Integer;
begin
  if Comp.FindForPlayer(Player) then
  begin
    i:=Comp.BestR;
    j:=Comp.BestC;
    k:=1;
    l:=1;
    while k <= Length(Comp.BestWord) do
    begin
      if Board.Squares[i, j].Empty then
      begin
        if l>5 then
        begin
          ShowMessage('PlayC '+IntToStr(l));
        end;
        if Comp.BestTiles[l]=0 then begin Inc(l); continue; end;

        Board.Squares[i, j].PlaceTile(Player.Squares[Comp.BestTiles[l]].Tile);
        Player.Squares[Comp.BestTiles[l]].RemoveTile;
        Inc(l);
      end;
      Inc(k);
      if Comp.BestDir then Inc(i) else Inc(j);
    end;
    Window.BoardDrawGrid.Invalidate;
    CurDrawGrid.Invalidate;
    PlaceWord;
  end
  else
  begin
    if Abort then exit;
    for i:=1 to 7 do Player.Squares[i].Mark;
    CurDrawGrid.Invalidate;
    if not ExchangeTiles then Pass;
  end;
end;

// *************************************************
// ******************** TPlayer ********************
// *************************************************

constructor TPlayer.Create(Game: TGame; Name: String; Num: Integer; Computer: Boolean=false);
var
  i: Integer;
begin
  inherited Create;
  self.Name:=Name;
  self.Num:=Num;
  self.Game:=Game;
  self.Computer:=Computer;
  Score:=0;
  TilesNum:=0;
  for i:=1 to 7 do Squares[i]:=TSquare.Create(nil, Self, 0, i, 1, 1);

  case Num of
    1: GroupBox:=Window.Player1GroupBox;
    2: GroupBox:=Window.Player2GroupBox;
    3: GroupBox:=Window.Player3GroupBox;
    4: GroupBox:=Window.Player4GroupBox;
    else GroupBox:=nil;
  end;

  for i:=0 to GroupBox.ControlCount-1 do
  if GroupBox.Controls[i].ClassName = 'TDrawGrid' then DrawGrid:=TDrawGrid(GroupBox.Controls[i])
  else if GroupBox.Controls[i].ClassName = 'TEdit' then ScoreEdit:=TEdit(GroupBox.Controls[i]);

  if not Computer then GroupBox.Caption:=Name
  else GroupBox.Caption:=Name+' - pota';
  DrawGrid.Visible:=false;
  GroupBox.Visible:=true;

  GetTiles;
  Draw;
end;

destructor TPlayer.Destroy;
var
  i: Integer;
begin
  GroupBox.Visible:=false;
  for i:=1 to 7 do Squares[i].Destroy;
  inherited Destroy;
end;

function TPlayer.GetTiles;
var
  i, c: Integer;
  tl: TList;
begin
  tl:=TList.Create;
  for i:=1 to 7 do if Squares[i].Tile <> nil then
  begin
    tl.Add(Squares[i].Tile);
    Squares[i].RemoveTile;
  end;

  if tl.Count=7 then c:=0
  else c:=Game.Bag.FetchTiles(7-tl.Count, tl);
  tl.Sort(SortTiles);
  Result:=c;

  for i:=1 to tl.Count do Squares[i].PlaceTile(tl[i-1]);

  TilesNum:=tl.Count;
  tl.Clear;
  tl.Free;
end;

procedure TPlayer.Draw;
var
  s: String;
begin
  Str(Score:4, s);
  ScoreEdit.Text:=s;
  GroupBox.Invalidate;
end;

procedure TPlayer.BackWord;
begin
  Game.Board.BackWord(Num);
end;

function TPlayer.BackTile(Tile: TTile): Boolean;
var
  i: Integer;
begin
  i:=1;
  Result:=false;
  while i <= 7 do
  if Squares[i].Empty then
  begin
    Result:=Squares[i].PlaceTile(Tile);
    break;
  end
  else Inc(i);
end;

procedure TPlayer.Activate;
begin
  if Game.Playing and not Computer then DrawGrid.Visible:=true;
  if Game.Playing then GroupBox.Font.Style:=[fsBold];
  Draw;
end;

procedure TPlayer.Deactivate;
begin
  GetTiles;
  DrawGrid.Visible:=false;
  if Game.Playing then GroupBox.Font.Style:=[];
  Draw;
end;

function TPlayer.ExchangeTiles: Boolean;
var
  i: Integer;
  tl: TList;
begin
  Result:=false;
  if Game.Board.Marked.Count=0 then exit;

  tl:=TList.Create;

  for i:=1 to 7 do
  if (Squares[i].Tile <> nil) and Squares[i].Marked then
  begin
    tl.Add(Squares[i].Tile);
    Squares[i].RemoveTile;
  end;

  Result:=Game.Bag.ExchangeTiles(tl);

  for i:=1 to 7 do if Squares[i].Tile <> nil then
  begin
    tl.Add(Squares[i].Tile);
    Squares[i].RemoveTile;
  end;

  tl.Sort(SortTiles);

  for i:=1 to tl.Count do Squares[i].PlaceTile(tl[i-1]);
  tl.Clear;
  tl.Free;
end;

// *************************************************
// ******************** TSquare ********************
// *************************************************

constructor TSquare.Create(Board: TBoard; Player: TPlayer; r, c, wm, lm: Integer);
begin
  inherited Create;

  Self.Board:=Board;
  Self.Player:=Player;
  Row:=r;
  Column:=c;
  WordMul:=wm;
  LetterMul:=lm;
  Empty:=true;
  ThisTurn:=false;
  Tile:=nil;
  Marked:=false;

  Bitmap:=TBitmap.Create;
  Window.TileImageList.GetBitmap(BitmapIndex, Bitmap);
end;

destructor TSquare.Destroy;
begin
  if Tile <> nil then Tile.Free;
  Bitmap.Free;
  inherited Destroy;
end;

function TSquare.BitmapIndex: Integer;
var r: Integer;
begin
  r:=BMP_BLANK;
  if Empty then
  begin
    if (Row=8) and (Column=8) then r:=BMP_STAR
    else if WordMul=3 then r:=BMP_W3
    else if WordMul=2 then r:=BMP_W2
    else if LetterMul=3 then r:=BMP_L3
    else if LetterMul=2 then r:=BMP_L2
    else r:=BMP_BLANK;
  end
  else if Marked then r:=BMP_CROSS
  else
  begin
    if Tile.Letter=' ' then r:=BMP_UNASGN
    else if Tile.Letter <= 'Z' then r:=ord(Tile.Letter)-ord('A')+BMP_LTR
    else case Tile.Letter of
      '': r:=BMP_CX;
      '': r:=BMP_DX;
      '': r:=BMP_EX;
      '': r:=BMP_NX;
      '': r:=BMP_RX;
      '': r:=BMP_SX;
      '': r:=BMP_TX;
      '': r:=BMP_UX;
      '': r:=BMP_ZX;
      '': r:=BMP_AY;
      '': r:=BMP_EY;
      '': r:=BMP_IY;
      '': r:=BMP_OY;
      '': r:=BMP_UY;
      '': r:=BMP_YY;
    end;
  end;
  Result:=r;
end;

function TSquare.PlaceTile(Tile: TTile): Boolean;
begin
  if not Empty then Result:=false
  else
  begin
    self.Tile:=Tile;
    Empty:=false;
    ThisTurn:=true;
    Result:=true;
    Window.TileImageList.GetBitmap(BitmapIndex, Bitmap);
    Marked:=false;

    if Board <> nil then Board.AddSquare(Self);
  end;
end;

procedure TSquare.RemoveTile(Method: Boolean=true);
begin
  Empty:=true;
  if (Tile <> nil) and Tile.Blank then Tile.Letter:=' ';
  Tile:=nil;
  ThisTurn:=false;
  Window.TileImageList.GetBitmap(BitmapIndex, Bitmap);
  Marked:=false;
  if Method and (Board <> nil) then Board.RemoveSquare(Self);
end;

function TSquare.Mark: Boolean;
begin
  Marked:=not Marked;
  Window.TileImageList.GetBitmap(BitmapIndex, Bitmap);
  Result:=Marked;
  Window.Game.Board.Mark(Self, Marked);
end;

function TSquare.AssignLetter(Letter: Char): Boolean;
begin
  if (Tile=nil) or not Tile.Blank then Result:=false
  else
  begin
    Tile.Letter:=Letter;
    Window.TileImageList.GetBitmap(BitmapIndex, Bitmap);
    Result:=true;
  end;
end;

function TPlayer.EndPlayer: Integer;
var
  i: Integer;
  c: Integer;
begin
  Result:=0;
  c:=0;
  for i:=1 to 7 do if Squares[i].Tile <> nil then
  begin
    Inc(c);
    Inc(Result, Squares[i].Tile.Value);
  end;
  Dec(Score, Result);
  if c=0 then Result:=-1;
end;

// ***********************************************
// ******************** TTile ********************
// ***********************************************

constructor TTile.Create(ltr: Char);
begin
  inherited Create;
  Letter:=ltr;
  Blank:=(ltr=' ');
end;

function TTile.Value: Integer;
begin
  if Blank then Result:=0
  else case Letter of
    'A': Result:=1;
    '': Result:=2;
    'B': Result:=3;
    'C': Result:=2;
    '': Result:=4;
    'D': Result:=1;
    '': Result:=8;
    'E': Result:=1;
    '': Result:=3;
    '': Result:=3;
    'F': Result:=5;
    'G': Result:=5;
    'H': Result:=2;
    'I': Result:=1;
    '': Result:=2;
    'J': Result:=2;
    'K': Result:=1;
    'L': Result:=1;
    'M': Result:=2;
    'N': Result:=1;
    '': Result:=6;
    'O': Result:=1;
    '': Result:=7;
    'P': Result:=1;
    'R': Result:=1;
    '': Result:=4;
    'S': Result:=1;
    '': Result:=4;
    'T': Result:=1;
    '': Result:=7;
    'U': Result:=2;
    '': Result:=5;
    '': Result:=4;
    'V': Result:=1;
    'X': Result:=10;
    'Y': Result:=2;
    '': Result:=4;
    'Z': Result:=2;
    '': Result:=4;
    else Result:=0;
  end;
end;

// ***********************************************
// ******************** TWord ********************
// ***********************************************

constructor TWord.Create(Text: String; Score: Integer);
begin
  inherited Create;
  Self.Text:=Text;
  Self.Score:=Score;
end;

function TWord.GetText: String;
var
  s: String;
  l: Integer;
begin
  l:=15-Length(Text);
  Str(Score:l, s);
  Result:=Text + s;
end;

// **********************************************
// ******************** TWindow ********************
// **********************************************

procedure TWindow.MenuExitClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TWindow.BoardDrawGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  lbl: String;
  r: TRect;
begin
  with BoardDrawGrid.Canvas do
  begin
    Brush.Color:=clGray;
    FillRect(Rect);
  end;

  if ARow = 0 then
  begin
    if ACol > 0 then with BoardDrawGrid.Canvas do
    begin
      lbl:=chr(ACol+ord('A')-1);
      TextOut(Rect.Left+(Rect.Right-Rect.Left-TextWidth(lbl)) div 2, Rect.Top+2, lbl);
    end
  end
  else if ACol = 0 then
  begin
    if ARow > 0 then with BoardDrawGrid.Canvas do
    begin
      Str(ARow, lbl);
      TextOut(Rect.Left+(Rect.Right-Rect.Left-TextWidth(lbl)) div 2, Rect.Top+2, lbl);
    end
  end
  else
  begin
    BoardDrawGrid.Canvas.Draw(Rect.Left, Rect.Top, Game.Board.Squares[ARow, ACol].Bitmap);
    if (Game.Board.Squares[ARow, ACol].Tile <> nil)
     and Game.Board.Squares[ARow, ACol].Tile.Blank then
    begin
      BoardDrawGrid.Canvas.Brush.Color:=clWhite;
      r.Left:=Rect.Left+21;
      r.Top:=Rect.Top+21;
      r.Right:=Rect.Right-1;
      r.Bottom:=Rect.Bottom-1;
      BoardDrawGrid.Canvas.FillRect(r);
    end;
    if Game.Board.LastTurn.IndexOf(Game.Board.Squares[ARow, ACol]) >=0 then
    begin
      BoardDrawGrid.Canvas.Brush.Color:=clRed;
      Inc(Rect.Left); Inc(Rect.Top);
      Dec(Rect.Right); Dec(Rect.Bottom);
      BoardDrawGrid.Canvas.FrameRect(Rect);
    end
    else if Game.Board.ThisTurn.IndexOf(Game.Board.Squares[ARow, ACol]) >=0 then
    begin
      BoardDrawGrid.Canvas.Brush.Color:=clBlue;
      Inc(Rect.Left); Inc(Rect.Top);
      Dec(Rect.Right); Dec(Rect.Bottom);
      BoardDrawGrid.Canvas.FrameRect(Rect);
    end;
  end;
end;

procedure TWindow.FormCreate(Sender: TObject);
begin
  Left:=(Screen.Width-Width) div 2;
  Top:=(Screen.Height-Height) div 2;

  Game:=TGame.Create;
  DragImage:=TDragImages.Create(Self);

  SetControls(Self);

//  Game.StartGame;
end;

procedure TWindow.BoardDrawGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  CanSelect:=false;
end;

procedure TWindow.PlayerDrawGridDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  pl: Integer;
begin
  pl:=Game.MapPlayerDrawGrid(Sender);
  if pl=0 then exit;
  Game.Players[pl].DrawGrid.Canvas.Draw(Rect.Left, Rect.Top, Game.Players[pl].Squares[ACol+1].Bitmap);
end;

procedure TWindow.BoardDrawGridDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept:=Source is TDragImages;
end;

procedure TWindow.PlayerDrawGridDragOver(Sender, Source: TObject; X,
  Y: Integer; State: TDragState; var Accept: Boolean);
begin
  Accept:=(Source is TDragImages) and ((Game.CurDrawGrid=Sender) or (Game.Board=Sender));
end;

procedure TWindow.PlayerDrawGridStartDrag(Sender: TObject;
  var DragObject: TDragObject);
begin
  DragObject:=DragImage;
end;

procedure TWindow.PlayerDrawGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  CanSelect:=false;
end;

procedure TWindow.BoardDrawGridDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var
  ACol, ARow: Integer;
begin
  if DragImage.Square=nil then exit;

  BoardDrawGrid.MouseToCell(X, Y, ACol, ARow);
  if (ARow=0) or (ACol=0) then exit;
  {
  if Board.Squares[ARow, ACol].PlaceTile(DragImage.Square.Tile, ' ') then
    if DragImage.Square.Row=0 then
    begin
      Players[curPlayer].Squares[DragImage.Square.Column].RemoveTile;
      BoardDrawGrid.Invalidate;
      Players[curPlayer].DrawGrid.Invalidate;
    end
    else
    begin
      Board.Squares[DragImage.Square.Row, DragImage.Square.Column].RemoveTile;
      BoardDrawGrid.Invalidate;
    end;
}
  if Game.Board.Squares[ARow, ACol].PlaceTile(DragImage.Square.Tile) then
  begin
    DragImage.Square.RemoveTile;
    DragImage.Square:=nil;
    BoardDrawGrid.Invalidate;
    Game.CurDrawGrid.Invalidate;
  end;
end;

procedure TWindow.BoardDrawGridStartDrag(Sender: TObject;
  var DragObject: TDragObject);
begin
  DragObject:=DragImage;
end;

procedure TWindow.BoardDrawGridMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ACol, ARow: Integer;
  l: Integer;
begin
  BoardDrawGrid.MouseToCell(X, Y, ACol, ARow);
  if (ARow=0) or (ACol=0) then exit;
  if Game.Board.Squares[ARow, ACol].Tile=nil then exit;
  if not Game.Board.Squares[ARow, ACol].ThisTurn then exit;

  if Button = mbLeft then
  begin
    DragImage.Square:=Game.Board.Squares[ARow, ACol];
    BoardDrawGrid.BeginDrag(false, 1);
  end
  else if Button = mbRight then
  begin
  if Game.Board.Squares[ARow, ACol].Tile.Blank then
  begin
    BlankTileForm.BlankTileLetterEdit.Text:=Game.Board.Squares[ARow, ACol].Tile.Letter;
    l:=BlankTileForm.ShowModal;
    if (l >= 32) then Game.Board.Squares[ARow, ACol].AssignLetter(chr(l));
    BoardDrawGrid.Invalidate;
    Game.Board.Draw;
  end;
  end;
end;

procedure TWindow.PlayerDrawGridMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  ACol, ARow: Integer;
begin
  if Game.CurDrawGrid <> Sender then exit;
  Game.CurDrawGrid.MouseToCell(X, Y, ACol, ARow);
  if Game.Player.Squares[ACol+1].Tile=nil then exit;
  if Button = mbLeft then
  begin
    if Game.Board.Marked.Count > 0 then exit;
    DragImage.Square:=Game.Player.Squares[ACol+1];
    Game.CurDrawGrid.BeginDrag(false, 1);
  end
  else if Button = mbRight then
  begin
    if Game.Board.ThisTurn.Count > 0 then exit;
    Game.Player.Squares[ACol+1].Mark;
    Game.CurDrawGrid.Invalidate;
  end;
end;

procedure TWindow.PlayerDrawGridDragDrop(Sender, Source: TObject; X,
  Y: Integer);
var
  ACol, ARow: Integer;
  pl: Integer;
begin
  if DragImage.Square=nil then exit;

  pl:=Game.MapPlayerDrawGrid(Sender);
  if pl=0 then exit;

  Game.Players[pl].DrawGrid.MouseToCell(X, Y, ACol, ARow);
  if Game.Players[pl].Squares[ACol+1].PlaceTile(DragImage.Square.Tile) then
  begin
    DragImage.Square.RemoveTile;
    DragImage.Square:=nil;
    BoardDrawGrid.Invalidate;
    Game.CurDrawGrid.Invalidate;
  end;
end;

procedure TWindow.BackWordButtonClick(Sender: TObject);
begin
  Game.Board.BackWord(Game.curPlayer);
end;

procedure TWindow.PassButtonClick(Sender: TObject);
begin
  Game.Pass;
end;

procedure TWindow.MenuNewGameClick(Sender: TObject);
begin
  if Game.Comp.Running then exit;
  Game.EndGame(END_RESET);
  Game.StartGame;
end;

procedure TWindow.MenuActionPlaceClick(Sender: TObject);
begin
  if PlaceWordButton.Enabled then PlaceWordButton.Click;
end;

procedure TWindow.MenuActionBackClick(Sender: TObject);
begin
  if BackWordButton.Enabled then BackWordButton.Click;
end;

procedure TWindow.MenuActionChallengeClick(Sender: TObject);
begin
  if ChallengeWordButton.Enabled then ChallengeWordButton.Click;
end;

procedure TWindow.MenuActionPassClick(Sender: TObject);
begin
  if PassButton.Enabled then PassButton.Click;
end;

procedure TWindow.MenuActionExchangeClick(Sender: TObject);
begin
  if ExchangeButton.Enabled then ExchangeButton.Click;
end;

procedure TWindow.OaplikaciClick(Sender: TObject);
begin
  Application.MessageBox('Jako zpotov program pro pedmt Programovn II (PRG031) v lt 2005 napsal Rastislav Wartiak.',
   'O aplikaci', MB_OK);
end;

procedure TWindow.PlaceWordButtonClick(Sender: TObject);
begin
  Game.PlaceWord;
end;

procedure TWindow.WordsStringGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  CanSelect:=false;
end;

procedure TWindow.ExchangeButtonClick(Sender: TObject);
begin
  Game.ExchangeTiles;
end;

procedure TWindow.ChallengeWordButtonClick(Sender: TObject);
begin
  Game.ChallengeWord;
end;

procedure TWindow.MenuEndGameClick(Sender: TObject);
begin
  Game.EndGame(END_RESET);
end;

procedure TWindow.HelpMeButtonClick(Sender: TObject);
begin
  Game.HelpPlayer;
end;

procedure TWindow.AbortButtonClick(Sender: TObject);
begin
  if Application.MessageBox('Ukonit hru?', 'Zastaven potae', MB_YESNO) = ID_YES then
  begin
    Game.Abort:=true;
    AbortButton.Enabled:=false;
  end;
end;

procedure TWindow.AddMemo(Line: String; Sep: Boolean = true);
begin
  WordsMemo.Lines.Add(Line);
  if Sep then WordsMemo.Lines.Add('   ---------------');
end;

procedure TWindow.MenuActionAbortClick(Sender: TObject);
begin
  if AbortButton.Enabled then AbortButton.Click;
end;

procedure TWindow.MenuActionHelpClick(Sender: TObject);
begin
  if HelpMeButton.Enabled then HelpMeButton.Click;
end;

end.
